﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Linq;
using Ascon.Pilot.SDK.ObjectsSample.ObjectCardView;
using Ascon.Pilot.Theme.Controls;
using Microsoft.Practices.Prism.Commands;

namespace Ascon.Pilot.SDK.ObjectsSample.NodeViewModel
{
    class ElementNodeViewModel : PropertyChangedBase, IObserver<IDataObject>
    {
        private string _displayName;
        private DataObjectWrapper _source;
        private readonly IObjectsRepository _repository;
        private readonly ITabServiceProvider _tabServiceProvider;
        private readonly IObjectModifier _modifier;
        private readonly IPilotDialogService _dialogService;
        private readonly IAttributeFormatParser _attributeFormatParser;
        private readonly ObservableCollection<ElementNodeViewModel> _childNodes;
        private readonly DelegateCommand _navigateToElementCommand;
        private readonly DelegateCommand _navigateToSourceFileCommand;
        private readonly DelegateCommand _showPropertiesCommand;
        private readonly DelegateCommand _deleteCommand;
        private bool _isExpanded;

        public ElementNodeViewModel()
        {
            DisplayName = "Loading";
        }

        public ElementNodeViewModel(IDataObject source, IObjectsRepository repository, ITabServiceProvider tabServiceProvider, IAttributeFormatParser attributeFormatParser, IPilotDialogService dialogService, IObjectModifier modifier)
        {
            if (dialogService == null) throw new ArgumentNullException("dialogService");

            _childNodes = new ObservableCollection<ElementNodeViewModel>();
            _source = new DataObjectWrapper(source, repository);
            _displayName = _source.DisplayName;
            _repository = repository;
            _tabServiceProvider = tabServiceProvider;
            _modifier = modifier;
            _dialogService = dialogService;
            _attributeFormatParser = attributeFormatParser;
            Id = source.Id;
            SourceFileId = source.RelatedSourceFiles.FirstOrDefault();

            _navigateToElementCommand = new DelegateCommand(NavigateToElement);
            _navigateToSourceFileCommand = new DelegateCommand(NavigateToSourceFile, CanNavigateToSourceFile);
            _showPropertiesCommand = new DelegateCommand(ShowProperties);
            _deleteCommand = new DelegateCommand(DeleteToRecycleBin);
        }

        public Guid Id { get; private set; }

        public Guid SourceFileId { get; private set; }

        public DelegateCommand NavigateToElementCommand
        {
            get { return _navigateToElementCommand; }
        }

        public DelegateCommand NavigateToSourceFileCommand
        {
            get { return _navigateToSourceFileCommand; }
        }

        public DelegateCommand ShowPropertiesCommand
        {
            get { return _showPropertiesCommand; }
        }

        public DelegateCommand DeleteCommand
        {
            get { return _deleteCommand; }
        }

        public string DisplayName
        {
            get { return _displayName; }
            set
            {
                _displayName = value;
                NotifyPropertyChanged("DisplayName");
            }
        }

        public ObservableCollection<ElementNodeViewModel> ChildNodes
        {
            get
            {
                if (_source == null)
                    return null;

                if (_source.ListViewChildren.Count == 0)
                    return _childNodes;

                if (_childNodes.Count != 0)
                    return _childNodes;

                _childNodes.Add(new LoadingElementNodeViewModel());
                return _childNodes;
            }
        }

        public bool IsExpanded
        {
            get { return _isExpanded; }
            set
            {
                if (_isExpanded == value)
                    return;

                if (value)
                {
                    if (_childNodes.Count == 1)
                    {
                        _repository.SubscribeObjects(_source.ListViewChildren).Subscribe(this);
                    }
                }

                _isExpanded = value;
                NotifyPropertyChanged("IsExpanded");
            }
        }

        public void OnNext(IDataObject value)
        {
            if (!_source.ListViewChildren.Contains(value.Id))
                return;

            if (IsLoading())
                _childNodes.Clear();

            if (!_childNodes.ToList().Exists(n => n.Id == value.Id))
                _childNodes.Add(new ElementNodeViewModel(value, _repository, _tabServiceProvider, _attributeFormatParser, _dialogService, _modifier));
            else
            {
                var node = _childNodes.First(n => n.Id == value.Id);
                node.Update(value);
            }
        }

        public void OnError(Exception error)
        {

        }

        public void OnCompleted()
        {

        }

        public void Update(IDataObject newSource)
        {
            _source = new DataObjectWrapper(newSource, _repository);
            DisplayName = _source.DisplayName;
            UpdateChildNodes();
        }

        private void UpdateChildNodes()
        {
            NotifyPropertyChanged("ChildNodes");
            RemoveDeletedNodes();
            LoadNewNodes();
        }

        private void RemoveDeletedNodes()
        {
            if (IsLoading())
                return;

            foreach (var childNode in _childNodes.ToList())
            {
                if (!_source.ListViewChildren.Contains(childNode.Id))
                    _childNodes.Remove(childNode);
            }
        }

        private void LoadNewNodes()
        {
            if (IsLoading())
                return;

            var childIdsToLoad = new List<Guid>();
            foreach (var childId in _source.ListViewChildren)
            {
                if (_childNodes.FirstOrDefault(x => x.Id == childId) == null)
                    childIdsToLoad.Add(childId);
            }

            _repository.SubscribeObjects(childIdsToLoad).Subscribe(this);
        }

        private bool IsLoading()
        {
            return _childNodes.FirstOrDefault() != null &&
                   _childNodes.First().GetType() == typeof(LoadingElementNodeViewModel);
        }

        private void NavigateToElement()
        {
            _tabServiceProvider.ShowElement(Id);
        }

        private void NavigateToSourceFile()
        {
            _tabServiceProvider.ShowElement(SourceFileId);
        }

        private bool CanNavigateToSourceFile()
        {
            return SourceFileId != Guid.Empty;
        }

        private void ShowProperties()
        {
            var objectCardViewModel = new ObjectCardViewModel(_source, _repository, _dialogService, _attributeFormatParser);
            var dialogView = new ObjectPropertiesView();
            dialogView.DataContext = objectCardViewModel;

            var propertiesWindow = new DialogWindow()
            {
                Title = "Edit element",
                Content = dialogView,
                Width = 400,
                Height = 450
            };

            propertiesWindow.ShowDialog();
        }

        private void DeleteToRecycleBin()
        {
            _modifier.DeleteById(Id);
            _modifier.Apply();
        }
    }

    class LoadingElementNodeViewModel : ElementNodeViewModel
    {
    }
}
