﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.ComponentModel;
using System.IO;
using System.Linq;
using System.Runtime.InteropServices;
using System.Windows;
using Ascon.Pilot.Theme.Tools;
using Microsoft.Win32;

namespace Ascon.Pilot.SDK.TaskSample.TaskEditView
{
    class TaskEditViewModel : INotifyPropertyChanged
    {
        protected readonly ITaskModifier _taskModifier;
        protected readonly IObjectsRepository _repository;
        private readonly IPilotDialogService _pilotDialogService;
        private readonly ITabServiceProvider _tabServiceProvider;
        protected ObservableCollection<TaskAttachment> _attachments = new ObservableCollection<TaskAttachment>();
        protected ObservableCollection<ExecutorItem> _executorItems = new ObservableCollection<ExecutorItem>();
        protected ObservableCollection<ExecutorItem> _auditorsItems = new ObservableCollection<ExecutorItem>();
        private string _title;
        private string _description;
        private bool _canSubmit;

        public TaskEditViewModel(ITaskModifier taskModifier, IObjectsRepository repository, IPilotDialogService pilotDialogService, ITabServiceProvider tabServiceProvider)
        {
            if (taskModifier == null)
                throw new ArgumentNullException("taskModifier");
            if (repository == null)
                throw new ArgumentNullException("repository");

            _taskModifier = taskModifier;
            _repository = repository;
            _pilotDialogService = pilotDialogService;
            _tabServiceProvider = tabServiceProvider;
            _executorItems.Add(new NewExecutorItem());
            _auditorsItems.Add(new NewAuditorItem());
        }

        public string Title
        {
            get { return _title; }
            set
            {
                _title = value;
                OnPropertyChanged("Title");
                UpdateCanSubmit();
            }
        }

        public string Description
        {
            get { return _description; }
            set
            {
                _description = value;
                OnPropertyChanged("Description");
                UpdateCanSubmit();
            }
        }

        public bool CanSubmit
        {
            get { return _canSubmit; }
            set
            {
                _canSubmit = value;
                OnPropertyChanged("CanSubmit");
            }
        }

        public ObservableCollection<TaskAttachment> Attachments
        {
            get { return _attachments; }
        }

        public ObservableCollection<ExecutorItem> ExecutorItems
        {
            get { return _executorItems; }
        }

        public ObservableCollection<ExecutorItem> AuditorsItems
        {
            get { return _auditorsItems; }
        }

        public void SelectExecutor()
        {
            var dialogOptions = _pilotDialogService
                .NewOptions()
                .WithParentWindow(GetActiveWindow())
                .WithCaption("Select executor")
                .WithOkButtonCaption("Accept")
                .WithAllowMultiSelect(AllowMultipleExecutors);

            var executors = _pilotDialogService.ShowPositionSelectorDialog(dialogOptions).Where(x => x.Children.Count == 0).ToList();
            if (!executors.Any())
                return;

            var people = _repository.GetPeople().ToList();

            _executorItems.Clear();
            foreach (var orgUnit in executors)
            {
                var person = people.FirstOrDefault(x => x.MainPosition != null && x.MainPosition.Position == orgUnit.Id);
                _executorItems.Add(new ExecutorItem(orgUnit.Id, person != null ? person.DisplayName : orgUnit.Title));
            }

            _executorItems.Add(new NewExecutorItem());
            UpdateCanSubmit();
        }

        public void SelectAuditor()
        {
            var dialogOptions = _pilotDialogService
                .NewOptions()
                .WithParentWindow(GetActiveWindow())
                .WithCaption("Select auditor")
                .WithOkButtonCaption("Accept")
                .WithAllowMultiSelect(AllowMultipleExecutors);

            var auditors = _pilotDialogService.ShowPositionSelectorDialog(dialogOptions).Where(x => x.Children.Count == 0).ToList();
            if (!auditors.Any())
                return;

            var people = _repository.GetPeople().ToList();

            _auditorsItems.Clear();
            foreach (var orgUnit in auditors)
            {
                var person = people.FirstOrDefault(x => x.MainPosition != null && x.MainPosition.Position == orgUnit.Id);
                _auditorsItems.Add(new ExecutorItem(orgUnit.Id, person != null ? person.DisplayName : orgUnit.Title));
            }

            _auditorsItems.Add(new NewAuditorItem());
            UpdateCanSubmit();
        }

        public void RemoveExecutor(int orgUnitId)
        {
            var executor = _executorItems.FirstOrDefault(x => x.OrgUnitId == orgUnitId);
            if (executor != null)
                _executorItems.Remove(executor);
            UpdateCanSubmit();
        }

        public void RemoveAttachment(Guid attachmentId)
        {
            var attachment = _attachments.FirstOrDefault(x => x.Id == attachmentId);
            if (attachment != null)
                _attachments.Remove(attachment);
            UpdateCanSubmit();
        }

        public void ShowAttachment(Guid attachmentId)
        {
            var attachment = _attachments.FirstOrDefault(x => x.Id == attachmentId);
            if (attachment == null)
                return;

            _tabServiceProvider.ShowElement(attachmentId);
        }

        public void AddAttachment()
        {
            var dialogOptions = _pilotDialogService
                .NewOptions()
                .WithParentWindow(GetActiveWindow())
                .WithCaption("Select attachments")
                .WithOkButtonCaption("Attach");

            var attachments = _pilotDialogService.ShowDocumentsSelectorDialog(dialogOptions);
            foreach (var attachment in attachments)
            {
                _attachments.Add(new TaskAttachment(attachment.Id, attachment.DisplayName, attachment.Type.Title, Theme.Icons.Instance.DocumentsIcon));
            }
            UpdateCanSubmit();
        }

        public void AddFileAttachment()
        {
            var storageItems = GetSelectedFiles();
            foreach (var attachment in storageItems)
            {
                _attachments.Add(new TaskAttachment(attachment.DataObject.Id, attachment.DataObject.DisplayName, "file", Theme.Icons.Instance.FileIcon));
            }
            UpdateCanSubmit();
        }

        public virtual bool SubmitTask()
        {
            if (!CanSubmit)
                return false;

            var executors = _executorItems.Where(x => x.OrgUnitId != -1).Select(x => x.OrgUnitId).ToList();
            var attachments = Attachments.Select(x => x.Id);
            var auditorIds = _auditorsItems.Where(a => a.OrgUnitId != -1).Select(a => a.OrgUnitId).ToList();
            var workFlow = _taskModifier.NewWorkflow().SetTitle(Title).SetDescription(Description).SetInitiatorAttachments(attachments.ToList());
            var stage = _taskModifier.NewStage(workFlow.WorkflowObject).SetStageOrder(1);
            foreach (var executorItemId in executors)
            {
                var buider = _taskModifier.NewTask(workFlow.WorkflowObject, stage.StageObject, executorItemId);
                buider.SetAuditors(auditorIds);
            }
            
            _taskModifier.Apply();

            return true;
        }

        public event PropertyChangedEventHandler PropertyChanged;

        protected virtual void OnPropertyChanged(string propertyName)
        {
            var handler = PropertyChanged;
            if (handler != null)
                handler(this, new PropertyChangedEventArgs(propertyName));
        }

        protected virtual void UpdateCanSubmit()
        {
            CanSubmit = (!string.IsNullOrEmpty(Title) || !string.IsNullOrEmpty(Description)) &&
                        _executorItems.Any(x => x.OrgUnitId != -1);
        }

        protected virtual bool AllowMultipleExecutors { get { return true; } }

        [DllImport("user32.dll")]
        public static extern IntPtr GetActiveWindow();

        private List<IStorageDataObject> GetSelectedFiles()
        {
            bool isValidPath;
            List<IStorageDataObject> storageItems;
            do
            {
                List<string> filePaths = null;
                string initialDirectory = _repository.GetStoragePath();
                var result = ShowOpenFileDialog("Select file attachments", null, true, initialDirectory, out filePaths);

                if (result != true)
                    return new List<IStorageDataObject>();

                storageItems = _repository.GetStorageObjects(filePaths).ToList();

                isValidPath = storageItems.All(x => x.State != StorageObjectState.None);

                if (!isValidPath)
                    MessageBox.Show("Only files from the pilot storage virtual drive can be attached.");

            } while (!isValidPath);

            return storageItems;
        }

        private bool? ShowOpenFileDialog(string title, string filter, bool mutiselect, string initialDirectory, out List<string> filePaths)
        {
            var dialog = new OpenFileDialog
            {
                Multiselect = mutiselect,
                Title = title
            };

            if (filter != null)
                dialog.Filter = filter;

            if (initialDirectory != null)
            {
                var fullPath = Path.GetFullPath(initialDirectory);
                dialog.InitialDirectory = fullPath;
            }

            try
            {
                return dialog.ShowDialog(LayoutHelper.GetOwnerWindow());
            }
            finally
            {
                filePaths = dialog.FileNames.ToList();
            }
        }
    }
}
