﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.ComponentModel.Composition;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Text.RegularExpressions;
using System.Windows;
using System.Windows.Media;
using System.Xml.Serialization;
using Ascon.Pilot.SDK.Toolbar;
using Ascon.Pilot.SDK.GraphicLayerSample;
using Ascon.Pilot.SDK.CreateObjectSample;
using Ascon.Pilot.SDK.ObjectsSample;
using Ascon.Pilot.Theme.ColorScheme;


namespace Ascon.Pilot.SDK.XpsViewerSample
{
    [Export(typeof(IToolbar<XpsRenderContext>))]
    public class XpsRenderToolbarSample : IToolbar<XpsRenderContext>, IObserver<INotification>
    {
        private const string TbiAdd = "tbiAddGraphicLayerElement";
        private const string TbiZoom = "tbiZoomGraphicLayerElement";
        private const string TbiZoomToAnnotation = "tbiZoomToAnnotation";

        private readonly IObjectModifier _modifier;
        private readonly IObjectsRepository _repository;
        private readonly IXpsViewer _xpsViewer;
        private readonly List<Guid> _graphicLayerQueueForZoom;
        private IPerson _currentPerson;
        private int _currentFileIndex;
        private int _currentAnnotationFileIndex;

        private List<FileWrapper> _objectFiles;

        [ImportingConstructor]
        public XpsRenderToolbarSample(IObjectModifier modifier, IObjectsRepository repository, IXpsViewer xpsViewer, IPilotDialogService dialogService)
        {
            _modifier = modifier;
            _repository = repository;
            _xpsViewer = xpsViewer;
            var convertFromString = ColorConverter.ConvertFromString(dialogService.AccentColor);
            _objectFiles = new List<FileWrapper>();
            _graphicLayerQueueForZoom = new List<Guid>();
            if (convertFromString != null)
            {
                var color = (Color)convertFromString;
                ColorScheme.Initialize(color, dialogService.Theme);
            }
            repository.SubscribeNotification(NotificationKind.ObjectFileChanged).Subscribe(this);
        }

        public void Build(IToolbarBuilder builder, XpsRenderContext context)
        {
            _objectFiles.Clear();
            _graphicLayerQueueForZoom.Clear();
            var zoomToAnnotation = "Перейти к аннотации";
            var addGraphicElement = "Добавить графический элемент";
            var editGraphicElement = "Показать графический элемент";
            builder.AddSeparator(builder.Count);
            builder.AddButtonItem(TbiZoomToAnnotation, builder.Count)
              .WithIcon(GetIcon("Icons/zoom_to_annotation_icon.svg"))
              .WithHeader(zoomToAnnotation)
              .WithShowHeader(false)
              .WithHint(zoomToAnnotation);
            builder.AddButtonItem(TbiAdd, builder.Count)
                .WithIcon(GetIcon("Icons/add_graphic_layer.svg"))
                .WithHeader(addGraphicElement)
                .WithShowHeader(false)
                .WithHint(addGraphicElement);
           
            builder.AddButtonItem(TbiZoom, builder.Count)
                .WithIcon(GetIcon("Icons/zoom_graphic_layer.svg"))
                .WithHeader(editGraphicElement)
                .WithShowHeader(false)
                .WithHint(editGraphicElement);
        }

        public void OnToolbarItemClick(string name, XpsRenderContext context)
        {
            _currentPerson = _repository.GetCurrentPerson();
            if (name == TbiZoomToAnnotation)
            {
                var loader = new ObjectLoader(_repository);
                loader.Load(context.DataObject.Id, obj =>
                {
                    var annotations = GetAnnotationFiles(context.DataObject, obj);
                    if (!annotations.Any())
                        return;

                    if (_currentAnnotationFileIndex > annotations.Count - 1)
                        _currentAnnotationFileIndex = 0;
                    var annotationFile = annotations.ElementAt(_currentAnnotationFileIndex);
                    if (annotationFile != null)
                    {
                        var fileName = annotationFile.Name;
                        var match = fileName.Split('_');
                        Guid guid;
                        if (match.Length > 1 && !string.IsNullOrEmpty(match[1]) && Guid.TryParse(match[1], out guid))
                            _xpsViewer.ZoomToElement(guid, 0.7);
                    }
                    _currentAnnotationFileIndex++;
                });
            }

            if (name == TbiAdd)
            {
                var loader = new ObjectLoader(_repository);
                loader.Load(context.DataObject.Id, AddGraphicLayerTextElementAndZoom);
            }

            if (name == TbiZoom)
            {
                var loader = new ObjectLoader(_repository);
                loader.Load(context.DataObject.Id, obj =>
                {
                    var graphicFiles = GetGraphicFilesFromCurrentVersion(context.DataObject, obj);
                    if (!graphicFiles.Any())
                        return;
                    
                    if (_currentFileIndex > graphicFiles.Count - 1)
                        _currentFileIndex = 0;
                    var currentGraphicLayerFile = graphicFiles.ElementAt(_currentFileIndex);
                    if (currentGraphicLayerFile != null)
                    {
                        var match = Regex.Match(currentGraphicLayerFile.Name, $"(?<={GraphicLayerElementConstants.GRAPHIC_LAYER_ELEMENT}).*?(?=_)");
                        Guid guid;
                        if (match.Success && Guid.TryParse(match.Value, out guid))
                            _xpsViewer.ZoomToElement(guid, 0.7);
                    }
                    _currentFileIndex++;
                });
            }
        }

        private List<IFile> GetGraphicFilesFromCurrentVersion(IDataObject contextObject, IDataObject loadedObject)
        {
            var currentVersionCreated = contextObject.ActualFileSnapshot.Created;
            var files = currentVersionCreated == loadedObject.ActualFileSnapshot.Created ? 
                loadedObject.Files : 
                loadedObject.PreviousFileSnapshots.First(s => s.Created == currentVersionCreated).Files;
            var graphicFiles = files.Where(f => f.Name.Contains(GraphicLayerElementConstants.GRAPHIC_LAYER_ELEMENT)).ToList();
            return graphicFiles;
        }

        private List<IFile> GetAnnotationFiles(IDataObject contextObject, IDataObject loadedObject)
        {
            var currentVersionCreated = contextObject.ActualFileSnapshot.Created;
            var files = currentVersionCreated == loadedObject.ActualFileSnapshot.Created ?
                loadedObject.Files :
                loadedObject.PreviousFileSnapshots.First(s => s.Created == currentVersionCreated).Files;
            var graphicFiles = files.Where(f => f.Name.Contains("Annotation")).ToList();
            return graphicFiles;
        }

        private void AddGraphicLayerTextElementAndZoom(IDataObject dataObject)
        {
            var elementId = Guid.NewGuid();
            var element = GraphicLayerElementCreator.CreateStampWithDateTime().ToString();
            SaveToDataBaseXaml(dataObject, element, elementId);
            _graphicLayerQueueForZoom.Add(elementId);
        }

        private void SaveToDataBaseXaml(IDataObject dataObject, string xamlObject, Guid elementId)
        {
            var builder = _modifier.Edit(dataObject);
            var positionId = _currentPerson.MainPosition.Position;
            var name = GraphicLayerElementConstants.GRAPHIC_LAYER_ELEMENT + elementId + "_" + positionId;
            var scale = new Point(1, 1);
            var element = GraphicLayerElementCreator.CreateFromFile(0, 0, scale, 0, positionId, VerticalAlignment.Top, HorizontalAlignment.Left, 
                GraphicLayerElementConstants.XAML, elementId, _xpsViewer.CurrentPageNumber);

            using (var stream = new MemoryStream())
            {
                var serializer = new XmlSerializer(typeof(GraphicLayerElement));
                serializer.Serialize(stream, element);
                stream.Position = 0;
                var e = serializer.Deserialize(stream);
                builder.AddFile(name, stream, DateTime.Now, DateTime.Now, DateTime.Now);
            }

            using (var textBlocksStream = new MemoryStream())
            {
                using (var writer = new StreamWriter(textBlocksStream))
                {
                    writer.Write(xamlObject);
                    writer.Flush();
                    builder.AddFile(GraphicLayerElementConstants.GRAPHIC_LAYER_ELEMENT_CONTENT + element.ContentId, textBlocksStream, DateTime.Now, DateTime.Now, DateTime.Now);
                }
            }
            _modifier.Apply();
        }

        private byte[] GetIcon(string relativePath)
        {
            var assembly = Assembly.GetEntryAssembly() ?? Assembly.GetExecutingAssembly();
            Application.ResourceAssembly = assembly;
            var uri = new Uri(relativePath, UriKind.Relative);
            var info = Application.GetResourceStream(uri);
            if (info == null)
                return null;

            using (var memoryStream = new MemoryStream())
            {
                info.Stream.CopyTo(memoryStream);
                return memoryStream.ToArray();
            }
        }

        public void OnNext(INotification notification)
        {
            _currentPerson = _repository.GetCurrentPerson();
            if (notification.ChangeKind == NotificationKind.ObjectFileChanged && _currentPerson.Id == notification.UserId)
            {
                var loader = new ObjectLoader(_repository);
                loader.Load(notification.ObjectId, obj =>
                {
                    var newFiles = obj.Files.Select(f => new FileWrapper(f)).ToList();
                    var addedFileNames = newFiles.Select(f => f.Name).Except(_objectFiles.Select(f => f.Name)).ToList();
                    foreach (var elementId in _graphicLayerQueueForZoom.ToList())
                    {
                        if (addedFileNames.Any(name => name.StartsWith(GraphicLayerElementConstants.GRAPHIC_LAYER_ELEMENT + elementId)))
                        {
                            _xpsViewer.ZoomToElement(elementId, 0.7);
                            _graphicLayerQueueForZoom.Remove(elementId);
                        }
                    }
                    _objectFiles = newFiles;
                });
            }
        }

        public void OnError(Exception error) { }
        public void OnCompleted() { }
    }
}