﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.ComponentModel;
using System.IO;
using System.Linq;
using System.Threading;
using System.Threading.Tasks;
using System.Windows.Threading;
using Ascon.Pilot.SDK;
using Ascon.Pilot.Theme;

namespace ChangeFilesMonitor
{
    class MainViewModel : PropertyChangedBase
    {
        private readonly IObjectsRepository _repository;
        private readonly Dispatcher _dispatcher;
        private FileItem _selectedItem;
        private readonly DelegateCommand.DelegateCommand _sendToServerCommand;
        private readonly DelegateCommand.DelegateCommand _sendAllToServerCommand;
        private readonly StorageCommandController _storageCommandController;
        private readonly DispatcherTimer _timer;
        private bool _isLoading;


        public MainViewModel(IObjectsRepository repository, Dispatcher dispatcher, StorageCommandController storageCommandController)
        {
            _repository = repository;
            _dispatcher = dispatcher;
            _storageCommandController = storageCommandController;

            ToolbarItemsSource = new ObservableCollection<ToolbarButtonItem>();
            FileItems = new ObservableCollection<FileItem>();

            _sendToServerCommand = new DelegateCommand.DelegateCommand(SendToServer, CanSend);
            _sendAllToServerCommand = new DelegateCommand.DelegateCommand(SendAllToServer, CanSendAll);

            InitializeToolbar();

            var root = GetRoot(repository);
            IsLoading = true;
            Load(root);

            _timer = new DispatcherTimer { Interval = new TimeSpan(0, 0, 0, 5) };
            _timer.Tick += (s, a) =>
            {
                Load(root);
            };
            _timer.Start();
        }


        public ObservableCollection<FileItem> FileItems { get; }
        public ObservableCollection<ToolbarButtonItem> ToolbarItemsSource { get; }

        public FileItem SelectedItem
        {
            get { return _selectedItem; }
            set
            {
                _selectedItem = value;
                NotifyOfPropertyChanged(() => SelectedItem);
                _sendToServerCommand.RaiseCanExecuteChanged();
            }
        }

        public bool IsLoading
        {
            get { return _isLoading; }
            set
            {
                _isLoading = value;
                NotifyOfPropertyChanged(() => IsLoading);
            }
        }

        private void SendAllToServer()
        {
            var items = FileItems.ToList();
            foreach (var fileItem in items)
            {
                SendToServer(fileItem);
            }
        }

        private bool CanSendAll()
        {
            return FileItems.Any();
        }

        private bool CanSend()
        {
            return SelectedItem != null;
        }

        private void SendToServer()
        {
            if (SelectedItem != null)
                SendToServer(SelectedItem);
        }

        private void SendToServer(FileItem item)
        {
            if (_storageCommandController.Commit(item.RelativPath))
            {
                FileItems.Remove(item);
            }
        }

        private void InitializeToolbar()
        {
            ToolbarItemsSource.Add(new ToolbarButtonItem("SendAll")
            {
                Header = "Отправить все",
                Icon = Icons.Instance.CopiedToTheActualVersionIcon,
                Command = _sendAllToServerCommand
            });

            ToolbarItemsSource.Add(new ToolbarButtonItem("SendSelected")
            {
                Header = "Отправить выделенный",
                Icon = Icons.Instance.ApplyRulesIcon,
                Command = _sendToServerCommand
            });
        }

        private string GetRoot(IObjectsRepository repository)
        {
            while (true)
            {
                var root = repository.GetStoragePath();
                if (!string.IsNullOrEmpty(root))
                    return root;

                Thread.Sleep(100);
            }
        }

        private async void Load(string root)
        {
            await LoadEditedFiles(root);
            IsLoading = false;
        }

        private async Task LoadEditedFiles(string path)
        {
            if (string.IsNullOrEmpty(path))
                return;

            await Task.Factory.StartNew(async () =>
            {
                var children = Directory.GetFileSystemEntries(path);
                foreach (var child in children)
                {
                    if (child.Contains("RecycleBin.lnk") || child.Contains("Корзина.lnk"))
                        continue;

                    var obj = _repository.GetStorageObjects(new List<string> { child });
                    var storageObj = obj.First();

                    if (storageObj.IsDirectory)
                    {
                        await LoadEditedFiles(child);
                        continue;
                    }

                    if (storageObj.State == StorageObjectState.Edited)
                        AddItem(storageObj.DataObject);
                    else
                        RemoveItem(storageObj.DataObject);
                }
            });
        }

        private async void RemoveItem(IDataObject obj)
        {
            await _dispatcher.BeginInvoke(new Action(() =>
            {
                var item = FileItems.FirstOrDefault(o => o.Id == obj.Id);
                if (item == null)
                    return;

                FileItems.Remove(item);
            }));
        }

        private async void AddItem(IDataObject obj)
        {
            await _dispatcher.BeginInvoke(new Action(() =>
            {
                var items = FileItems.ToList();
                if (items.Exists(i => i.Id == obj.Id))
                    return;

                var path = _repository.GetStoragePath(obj.Id);
                FileItems.Add(new FileItem(obj.Id, path, obj.DisplayName));
                //Подписываемся на изменения
                _repository.SubscribeObjects(new[] {obj.Id}).Subscribe(OnObjectChangeNext);
                _sendAllToServerCommand.RaiseCanExecuteChanged();
            }));
        }

        private void OnObjectChangeNext(IDataObject obj)
        {
            var item = FileItems.FirstOrDefault(o => o.Id == obj.Id);
            if (item == null)
                return;

            if (item.Name == obj.DisplayName)
                return;

            RemoveItem(obj);
            AddItem(obj);
        }
    }
}
