﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Reactive.Disposables;
using System.Reactive.Linq;
using DynamicData;
using DynamicData.Binding;
using DynamicData.Kernel;
using ProjectCloneWizard.Data;

namespace ProjectCloneWizard.ObjectsTree
{
    public class NodeViewModel : AbstractNotifyPropertyChanged, IDisposable
    {
        private readonly Node<DataObject, Guid> _node;
        private ReadOnlyObservableCollection<NodeViewModel> _children;
        private bool _isExpanded;
        private readonly IDisposable _cleanUp;
        private bool _isChecked = true;

        public NodeViewModel(Node<DataObject, Guid> node, IDataObjectService dataObjectService, bool isExpand, NodeViewModel parent = null)
        {
            _node = node;
            Parent = parent;

            if (Parent.HasValue)
                IsChecked = Parent.Value.IsChecked;

            //Документы не выбираем изначально
            if (node.Item.Type.Children.Count == 0 && node.Item.Type.HasFiles && !node.Item.Type.IsEcmDocument)
                _isChecked = false;

            var childrenLoader = new Lazy<IDisposable>(() => node.Children.Connect()
                    .Transform(x => new NodeViewModel(x, dataObjectService, isExpand, this))
                    .Sort(new NodeViewModelComparer())
                    .Bind(out _children)
                    .DisposeMany()
                    .Subscribe());

            var shouldExpand = node.IsRoot
                ? Observable.Return(true)
                : Parent.Value.WhenValueChanged(x => x.IsExpanded);

            if (isExpand)
                shouldExpand = Observable.Return(true);

            var expander = shouldExpand
                    .Where(isExpanded => isExpanded)
                    .Take(1)
                    .Subscribe(x =>
                    {
                        dataObjectService.Load(node.Item.Id, () => { });
                        //force lazy loading
                        var value = childrenLoader.Value;
                    });

            _cleanUp = Disposable.Create(() =>
            {
                expander.Dispose();
                if (childrenLoader.IsValueCreated)
                    childrenLoader.Value.Dispose();
            });
        }

        public Guid Id => _node.Item.Id;
        public string Title => _node.Item.Title;
        public byte[] Icon => _node.Item.Icon;
        public int Sort => _node.Item.Sort;
        public Optional<NodeViewModel> Parent { get; }
        public ReadOnlyObservableCollection<NodeViewModel> Children => _children;
        public DataType Type => _node.Item.Type;
        public IDictionary<string, object> Attributes => _node.Item.Attributes;
        public ReadOnlyCollection<AccessRecord> Access => _node.Item.Access;
        public bool IsSecret => _node.Item.IsSecret;

        public bool IsExpanded
        {
            get { return _isExpanded; }
            set { SetAndRaise(ref _isExpanded, value); }
        }

        public bool IsChecked
        {
            get { return _isChecked; }
            set
            {
                if (value && Parent.HasValue)
                    Parent.Value.IsChecked = true;
                
                if (!value && Children != null)
                {
                    foreach (var child in Children)
                        child.IsChecked = false;
                }

                SetAndRaise(ref _isChecked, value);
            }
        }

        public void Dispose()
        {
            _cleanUp.Dispose();
        }
    }
}
